@@ -311,6 +311,7 @@ static void WeakPointerCallback(v8::Persistent<v8::Value> handle, void* id) {
 }
 
 TEST(ObjectGroups) {
+  FLAG_incremental_marking = false;
   InitializeVM();
   GlobalHandles* global_handles = Isolate::Current()->global_handles();
 
@@ -545,9 +546,9 @@ TEST(BootUpMemoryUse) {
       }
     } else {
       if (v8::internal::Snapshot::IsEnabled()) {
-        CHECK_LE(delta, 2600 * 1024);  // 2484.
+        CHECK_LE(delta, 2500 * 1024);  // 2400.
       } else {
-        CHECK_LE(delta, 2950 * 1024);  // 2844
+        CHECK_LE(delta, 2860 * 1024);  // 2760.
       }
     }
   }
